import torch
from jaxtyping import Float
from torch import Tensor

from bandai.utils import batched_quadratic_form

from .protocols import ArmDist, Context, ContextualMAB, Rewards

ArmMat = Float[Tensor, " A D D"]
ArmVec = Float[Tensor, " A D"]


class DisjointLinUCB(ContextualMAB):
    """
    DisjointLinUCB.
    """

    def __init__(
        self,
        n_arms: int,
        ctx_dim: int,
        exploration_coef: float = 1.0,
    ) -> None:
        self._alpha = exploration_coef

        self._cov_mat: ArmMat = torch.eye(ctx_dim).repeat(n_arms, 1, 1)
        self._inv_cov: ArmMat = torch.eye(ctx_dim).repeat(n_arms, 1, 1)
        self._bias: ArmVec = torch.zeros(n_arms, ctx_dim)
        self._estim: ArmVec = torch.zeros(n_arms, ctx_dim)

    def update(self, arm: int, ctx: Context, rwd: Rewards) -> None:
        # update A = A + X'X
        self._cov_mat[arm] += ctx.T @ ctx

        # update b = b + X'r
        self._bias[arm] += ctx.T @ rwd

        # recompute inv(A)
        self._inv_cov[arm] = torch.linalg.inv(self._cov_mat[arm])  # type: ignore

        # update estimate \theta = inv(A) @ b
        self._estim[arm] = self._inv_cov[arm] @ self._bias[arm]

    def predict(self, ctx: Context) -> ArmDist:
        # compute mean estimtate x'\theta
        mean = ctx @ self._estim.T

        # compute upper confidence bound sqrt(x'Ax)
        var = batched_quadratic_form(ctx, self._inv_cov)

        # return final UCB scores
        return mean + self._alpha * var.sqrt()



import pytest
import torch
from torch import Tensor

from bandai.utils import batched_quadratic_form


@pytest.fixture
def rand_vec_batch() -> Tensor:
    """
    Random batch of vectors.
    """
    batch, dim = 2, 3
    return torch.randn(batch, dim)


@pytest.fixture
def rand_symmat_batch() -> Tensor:
    """
    Random batch of symmetric matrices.
    """
    batch, dim = 2, 3
    mat = torch.randn(batch, dim, dim)
    return 0.5 * (mat + mat.transpose(-1, -2))  # symmetrize


def manual_quadratic_form(vec: Tensor, mat: Tensor) -> Tensor:
    """
    Compute x'Ax manually for all combinations of x from vec and A from mat.
    """
    vec_batch, mat_batch = vec.shape[0], mat.shape[0]
    result = torch.zeros(vec_batch, mat_batch)

    for i in range(vec_batch):
        for j in range(mat_batch):
            result[i, j] = vec[i] @ mat[j] @ vec[i]
    return result


def test_batched_quadratic_form(rand_vec_batch: Tensor, rand_symmat_batch: Tensor) -> None:
    expected = manual_quadratic_form(rand_vec_batch, rand_symmat_batch)
    result = batched_quadratic_form(rand_vec_batch, rand_symmat_batch)

    assert result.shape == expected.shape
    assert torch.allclose(result, expected, atol=1e-5)



import subprocess


def test_hello_from_bandai() -> None:
    result = subprocess.run(
        [
            "bandai",
        ],
        capture_output=True,
        text=True,
    )

    expected = """Hello from bandai!"""

    assert result.returncode == 0
    assert result.stdout.strip() == expected.strip()



from typing import Protocol

from jaxtyping import Float
from torch import Tensor

ArmDist = Float[Tensor, " B A"]
Context = Float[Tensor, " B D"]
Rewards = Float[Tensor, " B"]


class ContextualMAB(Protocol):
    """
    Contextual Multi-Armed Bandit Protocol.
    """

    def update(self, arm: int, ctx: Context, rwd: Rewards) -> None: ...

    def predict(self, ctx: Context) -> ArmDist: ...



import torch
from jaxtyping import Float
from torch import Tensor


def batched_quadratic_form(
    batch_vec: Float[Tensor, " B D"],
    batch_mat: Float[Tensor, " A D D"],
) -> Float[Tensor, " B A"]:
    """
    Computes batched quadratic form action.
    """
    return torch.einsum("bd, acd, bc -> ba", batch_vec, batch_mat, batch_vec)
